#include <mpi.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "mandel.h"

int rank,size;
char **Image;
const int imgWidth=4096;
const int imgHeight=4096;
char Row[imgWidth];

void Master()
{
    
	MPI_Status Status;
	
	int Slaves=size-1;
	for (int i=1;i<size;i++) {
		int Slave=i-1;
		int NRows=imgHeight/Slaves;
		int StartRow=Slave*NRows;
		int Remainder=imgHeight%Slaves;
		if (Slave<Remainder) {
			NRows++;
			StartRow+=Slave;
		} else
			StartRow+=Remainder;
		for(int j=0;j<NRows;j++) {
			MPI_Recv(Row,imgWidth,MPI_CHAR,i,MPI_ANY_TAG,MPI_COMM_WORLD,&Status);
			int ReceivedRow=Status.MPI_TAG;
			memcpy(Image[ReceivedRow],Row,sizeof(char)*imgWidth);
		}
	}
}

void Slave()
{
	int Slaves=size-1;
	int Slave=rank-1;
	int NRows=imgHeight/Slaves;
	int StartRow=Slave*NRows;
	int Remainder=imgHeight%Slaves;
	if (Slave<Remainder) {
		NRows++;
		StartRow+=Slave;
	} else
		StartRow+=Remainder;
	
	for(int i=StartRow;i<StartRow+NRows;i++) {
		CalcImgRow(Row,i,imgWidth,imgHeight);
		MPI_Send(Row,imgWidth,MPI_CHAR,0,i,MPI_COMM_WORLD);
	}
}

int main(int argc,char *argv[])
{
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,&rank);
	MPI_Comm_size(MPI_COMM_WORLD,&size);
	if (size<2) {
		printf("At least 2 processes are needed\n");
		MPI_Finalize();
		exit(-1);
	}
	MPI_Barrier(MPI_COMM_WORLD);
	if (rank==0) { 
		AllocMem(imgWidth,imgHeight);
		double t1=MPI_Wtime();
		Master();
		double t2=MPI_Wtime();
		printf("Execution time: %f seconds\n",t2-t1);
		WritePGM("mandel.pgm",imgWidth,imgHeight);
		FreeMem(imgWidth,imgHeight);
	} else
		Slave();	
	MPI_Finalize();
}
